# Compilation
Using `cmake`, the project should compile without any further dependencies:
```
cmake .
make 
```

# Usage
Compilation produces two binaries (`dejavu-auto` and `dejavu-iso`), which are similar in usage. `dejavu-auto` is the binary of the automorphism solver, `dejavu-iso` of the isomorphism solver. For interesting benchmark graphs you may check out the [nauty / Traces benchmark library](https://pallini.di.uniroma1.it/Graphs.html).

## dejavu-auto
The automorphism solver `dejavu-auto` only accepts files in the [DIMACS graph format](http://www.tcs.hut.fi/Software/bliss/fileformat.shtml). Only undirected graphs can be handled at this point -- but the graphs may be colored. An example call using 8 threads might be:
```
dejavu_auto pp-25-100 --threads 7
```

The solver accepts the following command line arguments:

Command Line Argument | Effect
--- | ---
`--threads` | specifies the amount of additional threads used (total is threads + 1, default is 0)
`--write-auto` | outputs a generating set of the found automorphism group
`--timeout` | specifies a timeout (in seconds)
`--permute` | permutes the input graph randomly
`--permute-seed` | specify a fixed seed for the previous argument
`--force-selector` | forces a specific cell selector (0 is first, 1 is smallest, 2 is largest)
`--no-idleskip` | suppresses skipping of idle cells, i.e., cells that cause splits
`--kdeviation` | determine the number of additional refinement steps for deviation values

## dejavu-iso
The isomorphism solver `dejavu-iso` mainly accepts the same arguments, however, it requires two input files. An example call using 8 threads might be:
```
dejavu_iso pp-25-100 pp-25-101 --threads 7
```

The solver accepts the following command line arguments:

Command Line Argument | Effect
--- | ---
`--threads` | specifies the amount of additional threads used (total is threads + 1, default is 0)
`--write-auto` | outputs a generating set of the found automorphism group
`--timeout` | specifies a timeout (in seconds)
`--permute` | permutes the input graph randomly
`--permute-seed` | specify a fixed seed for the previous argument
`--force-selector` | forces a specific cell selector (0 is first, 1 is smallest, 2 is largest)
`--no-idleskip` | suppresses skipping of idle cells, i.e., cells that cause splits
`--kdeviation` | determine the number of additional refinement steps for deviation values

# Work in Progress
A new version, which in particular includes a much more user-friendly API, is in the works. All feedback is appreciated!

# License
The source code contains modified source code of the [nauty / Traces](http://pallini.di.uniroma1.it) distribution, as well as the lock-free queue implementation described [here](http://moodycamel.com/blog/2014/a-fast-general-purpose-lock-free-queue-for-c++). 

1. All files not listed below, copyright Markus Anders, MIT license.
2. `naudefs.h`, `naurng.cpp`, `naurng.h`, `schreier_sequential.cpp`, `schreier_sequential.h`, `schreier_shared.cpp`, `schreier_shared.h` are (derivative) work, originally copyright Brendan McKay. Licensed under the Apache License, Version 2.0 (see `LICENSE`).
3. `concurrentqueue.h`, copyright Cameron Desrochers. Licensed under the Simplified BSD license (see `concurrentqueue.h`)
